# metadata, base, engine and models used in testing

from sqlalchemy import (
    engine_from_config, Column,
    ForeignKey,
    Integer, Text, DateTime, Interval, BLOB
)
from py_liant.enum import PythonEnum
from sqlalchemy.orm import (
    sessionmaker, configure_mappers,
    relationship, backref
)
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.schema import MetaData
import zope.sqlalchemy
from enum import Enum


class BaseBase:
    # dynamic table name should be OK for testing

    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()


# model definition
metadata = MetaData()
Base = declarative_base(metadata=metadata, cls=BaseBase)


class ParentType(Enum):
    type1 = 'type1'
    type2 = 'type2'
    type3 = 'type3'


class Parent(Base):
    id = Column(Integer, primary_key=True)
    data1 = Column(Text)
    data2 = Column(DateTime)
    data3 = Column(Interval)
    data4 = Column(BLOB)
    data5 = Column(PythonEnum(ParentType))


class Child(Base):
    id = Column(Integer, primary_key=True)
    parent_id = Column(ForeignKey(Parent.id, ondelete="CASCADE"), index=True)
    data = Column(Text)

    parent = relationship(Parent, backref=backref(
        'children', cascade='all, delete-orphan'))


# finalize mappers
configure_mappers()


def get_engine(settings, prefix='sqlalchemy.'):
    return engine_from_config(settings, prefix)


def get_session_factory(engine):
    factory = sessionmaker()
    factory.configure(bind=engine)
    return factory


def get_tm_session(session_factory, transaction_manager):
    """
    Get a ``sqlalchemy.orm.Session`` instance backed by a transaction.

    This function will hook the session to the transaction manager which
    will take care of committing any changes.

    - When using pyramid_tm it will automatically be committed or aborted
      depending on whether an exception is raised.

    - When using scripts you should wrap the session in a manager yourself.
      For example::

          import transaction

          engine = get_engine(settings)
          session_factory = get_session_factory(engine)
          with transaction.manager:
              dbsession = get_tm_session(session_factory, transaction.manager)

    """
    dbsession = session_factory()
    zope.sqlalchemy.register(
        dbsession, transaction_manager=transaction_manager)
    return dbsession


def includeme(config):
    """
    Initialize the model for a Pyramid app.

    Activate this setup using ``config.include('cumulus_auth.models')``.

    """
    settings = config.get_settings()
    settings['tm.manager_hook'] = 'pyramid_tm.explicit_manager'

    # use pyramid_tm to hook the transaction lifecycle to the request
    config.include('pyramid_tm')

    # use pyramid_retry to retry a request when transient exceptions occur
    # config.include('pyramid_retry')

    session_factory = get_session_factory(get_engine(settings))
    config.registry['dbsession_factory'] = session_factory

    # make request.dbsession available for use in Pyramid
    config.add_request_method(
        # r.tm is the transaction manager used by pyramid_tm
        lambda r: get_tm_session(session_factory, r.tm),
        'dbsession',
        reify=True
    )



from collections import OrderedDict


class JsonObject(dict):

    def __getattr__(self, item):
        if item not in self:
            raise AttributeError("No such attribute: " + item)
        return self[item]

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, item):
        if item in self:
            del self[item]
        else:
            raise AttributeError("No such attribute: " + item)

    def __hash__(self):
        return id(self)


class JsonOrderedObject(OrderedDict):
    # same as JsonObject but based on OrderedDict, useful to get predictable
    # results during testing

    def __getattr__(self, item):
        if item not in self:
            raise AttributeError("No such attribute: " + item)
        return self[item]

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, item):
        if item in self:
            del self[item]
        else:
            raise AttributeError("No such attribute: " + item)

    def __hash__(self):
        return id(self)



from pyparsing import (Group, oneOf, Word, alphas, alphanums, nums,
                       delimitedList, Literal, Forward, Optional, Suppress
                       )

field_chars = alphanums + '_'
fieldname = Word(alphas, field_chars)

# a field -- value property to be ignored (-)
field = Group(Literal('-')('op') + fieldname('name'))
# a relationship, collection or single value, to be loaded
collection = Forward()
# a casting directive;
caster = Forward()
hints_parser = delimitedList(field | collection | caster)
hints_parser_nocast = delimitedList(field | collection)
collection << Group(oneOf('+ *')('op') + fieldname('name') +
                    Optional(Suppress('(') + hints_parser +
                             Suppress(')'))('children'))
caster << Group(Literal("!")('op') + fieldname('type') +
                Suppress('(') + hints_parser_nocast('children') +
                Suppress(')')
                )

# route syntax is: verb@pkey,pkey/drilldown[slice]:hints

pkey_chars = alphanums + '_- '

integer_number = Word('-'+nums, nums)

index_parser = Group(
    integer_number('start') + Suppress(':') + integer_number('stop')
    | integer_number('index'))


route_parser = fieldname('verb') + \
    Optional(Suppress('!') + fieldname('cast')) + \
    Optional(Suppress('@') + delimitedList(Word(pkey_chars))('pkey')) + \
    Optional(Suppress('/') + fieldname('drilldown') +
             Optional(Suppress('!') + fieldname('drilldown_cast'))) + \
    Optional(Suppress('[') + index_parser('slice') + Suppress(']')) + \
    Optional(Suppress(':') + fieldname('profile')) + \
    Optional(Suppress(':') + hints_parser('hints'))



from .json_object import JsonObject
from .json_decoder import JSONDecoder
from .json_encoder import JSONEncoder
from .monkeypatch import patch_sqlalchemy_base_class
from .pyramid import (
    CRUDView, pyramid_json_renderer_factory, pyramid_json_decoder)

__all__ = ["JsonObject", "JSONDecoder", "JSONEncoder",
           "patch_sqlalchemy_base_class", "CRUDView",
           "pyramid_json_renderer_factory", "pyramid_json_decoder"]


# example usage as pyramid configuration function
# def includeme(config):
#     config.add_renderer('json', pyramid_json_renderer_factory(Base))
#     config.add_request_method(pyramid_json_decoder, 'json', reify=True)
#     patch_sqlalchemy_base_class(Base)



import base64
import uuid
from collections import OrderedDict
from datetime import date, datetime, time
from decimal import Decimal
from enum import Enum

from dateutil import parser, tz
from sqlalchemy import Column, DateTime, String, Time, cast
from sqlalchemy.dialects.postgresql import HSTORE, UUID
from sqlalchemy.ext.associationproxy import ASSOCIATION_PROXY
from sqlalchemy.ext.hybrid import HYBRID_PROPERTY
from sqlalchemy.inspection import inspect
from sqlalchemy.orm import (ColumnProperty, CompositeProperty, Mapper,
                            RelationshipProperty, Session, SynonymProperty, column_property)
from sqlalchemy.types import ARRAY

from pyramid.settings import asbool

from .interfaces import JsonGuardProvider


def coerce_value(cls, column, value, size_check=True):
    if value is None:
        if not column.nullable:
            raise ValueError(
                f'Null value not allowed for property {column.key} of type '
                f'{cls!r}')
        return None

    try:
        python_type = column.type.python_type
    except NotImplementedError:
        # SA types that don't implement python_type treated early and explicity
        if column.type is HSTORE or type(column.type) is HSTORE:
            if value is None or isinstance(value, dict):
                # coerce all non-string values of hstore dict
                return {key: str(val) for key, val in value.items()}
            raise ValueError(
                f'Invalid type {type(value)!r} for property {column.key} of '
                f'class {cls!r}')
        if column.type is UUID or type(column.type) is UUID:
            if column.type.as_uuid:
                if type(value) is int:
                    return uuid.UUID(int=value)
                return uuid.UUID(value)
            return str(value)
        raise NotImplementedError(f'py_liant does not support {column.type!r} '
                                  'yet')

    if python_type is str:
        if value is not str:
            value = str(value)
        if isinstance(column.type, String) and hasattr(column.type, "length") \
                and value is not None and column.type.length is not None \
                and column.type.length < len(value) and size_check:
            raise ValueError(
                f'Text value too large for property {column.key} of class '
                f'{cls!r}, limit {column.type.length}')

    if python_type is int:
        try:
            value = int(value)
        except ValueError:
            raise ValueError(
                'Could not convert value to target type for property '
                f'{column.key} of class {cls!r}')

    if python_type is Decimal and value is not None \
            and type(value) is not Decimal:
        try:
            value = Decimal(value)
        except ArithmeticError:
            raise ValueError(
                'Could not convert value to target type for property '
                f'{column.key} of class {cls!r}')

    if python_type is float and value is not None and type(value) is not float:
        try:
            value = float(value)
        except ValueError:
            raise ValueError(
                'Could not convert value to target type for property '
                f'{column.key} of class {cls!r}')

    if python_type is bool and value is not None and type(value) is not bool:
        try:
            if type(value) is str:
                value = asbool(value)
            elif type(value) is int:
                if value not in [0, 1]:
                    raise ValueError
                value = bool(value)
            else:
                raise ValueError(
                    'Could not convert value to target type for property '
                    f'{column.key} of class {cls!r}')
        except ValueError:
            raise ValueError(
                f'Expected boolean value for property {column.key} of class '
                f'{cls!r}, received {type(value)!r} instead')

    if python_type in (datetime, date, time):
        try:
            use_timezone = False
            if isinstance(column.type, (DateTime, Time)):
                use_timezone = column.type.timezone

            def tzinfos(name, offset):
                if offset is not None:
                    return tz.tzoffset(name, offset)
                return tz.gettz(name) or None

            value = parser.parse(value, tzinfos=tzinfos,
                                 ignoretz=not use_timezone)
            if python_type is date:
                value = value.date()

            if python_type is time:
                value = value.timetz() if use_timezone else value.time()
        except ValueError:
            raise ValueError(
                'Could not convert value to target type for property '
                f'{column.key} of class {cls!r}')

    if python_type is bytes:
        return base64.b64decode(value)
    
    if python_type is list:
        if column.type is ARRAY or isinstance(column.type, ARRAY):
            # use SQL casting for arrays
            return cast(value, column.type)
        raise NotImplementedError(f'py_liant does not support {column.type!r} '
                                  'yet')

    if issubclass(python_type, Enum) and value is not None:
        value = str(value)
        try:
            value = next(item for item in python_type
                         if item.name == value or item.value == value)
        except StopIteration:
            raise ValueError(
                'Could not convert value to target type for property '
                f'{column.key} of class {cls!r}')

    return value


def _get_pk_from_json(cls, pk_tuple, child_data, fk_pairs=None, parent=None):
    ret = list()

    for col in pk_tuple:
        if col.key in child_data:
            value = child_data[col.key]
            value = coerce_value(cls, col, value)
            ret.append(value)
            continue
        if fk_pairs is not None and parent is not None:
            parent_col = next((pair[0]
                               for pair in fk_pairs if pair[1] == col), None)
            if parent_col is not None:
                parent_mapper = inspect(type(parent))
                value = getattr(parent, parent_mapper.get_property_by_column(
                    parent_col).class_attribute.key)
                ret.append(value)
                continue
        # a foreign key column's value could not be found - return None
        # (it's up to the caller to determine what this means)
        ret.append(None)
    return ret


def _polymorphic_constructor(cls, data):
    polymorphic_col = cls.__mapper__.polymorphic_on
    if polymorphic_col is None:
        return cls()
    polymorphic_prop = cls.__mapper__.get_property_by_column(polymorphic_col)
    if polymorphic_prop.key not in data:
        # assume this was intentional
        return cls()
    identity = coerce_value(cls, polymorphic_col, data[polymorphic_prop.key])
    return cls.__mapper__.polymorphic_map[identity].class_()


def __apply_changes(self, data, object_dict=None, context=None,
                    for_update=True):
    if object_dict is None:
        object_dict = dict()

    # register in object dictionary to prevent loops and worse
    if self in object_dict:
        return
    object_dict[data] = self
    mapper = inspect(type(self))

    if isinstance(context, JsonGuardProvider):
        if set(data.keys()) - {col.name for col in mapper.primary_key}:
            if not context.guardUpdate(self, data, for_update):
                return

    for key, attr in mapper.all_orm_descriptors.items():
        if key not in data:
            continue

        value = data[key]

        if attr.extension_type == HYBRID_PROPERTY:
            # for hybrid properties we cannot rely on any typing system;
            # all checks and conversions if any are needed should be done by
            # the setter method
            setter = getattr(attr, 'fset', None)
            if setter is None:
                continue
            setter(self, value)
        elif attr.is_attribute and hasattr(attr, 'property'):
            prop = attr.property
            if isinstance(prop, ColumnProperty):
                # guaranteed there is only one column in the property?
                assert len(prop.columns) == 1
                column = prop.columns[0]
                if not isinstance(column, Column):
                    continue

                # updating a primary key column with autoincrement: no
                if column in (table._autoincrement_column
                              for table in mapper.tables
                              if table._autoincrement_column is not None):
                    continue

                value = coerce_value(type(self), column, value)
                attr.__set__(self, value)
            elif isinstance(prop, CompositeProperty):
                # composite properties are also quite hard to get right;
                # should rely either on constructor or on value coercion
                attr.__set__(self, value)
            elif isinstance(prop, SynonymProperty):
                raise NotImplementedError('synonym properties not supported')
            elif isinstance(prop, RelationshipProperty):
                __apply_collection_changes(self, attr, value, object_dict,
                                           context)
            else:
                raise AssertionError(
                    f'Unexpected property type {type(prop)!r} in '
                    f'{type(self)!r}: {key}')
        elif attr.extension_type == ASSOCIATION_PROXY:
            raise NotImplementedError(
                'association proxy updates not supported')
        else:
            raise AssertionError(
                f'Unexpected extension type {attr.extension_type!r} '
                f'in {type(self)!r}: {key}')


def __apply_collection_changes(self, attr, value, object_dict, context):
    prop = attr.property
    child_class = prop.argument

    if isinstance(child_class, Mapper):
        child_class = child_class.class_

    child_mapper = inspect(child_class, False)

    # might need to resolve (class resolver instance)
    if child_mapper is None:
        child_class = child_class()
        child_mapper = inspect(child_class)

    if prop.uselist:
        # list of items
        # TODO implement mapped collections?
        collection = attr.__get__(self, None)

        # empty list/dict or null
        if not value:
            collection.clear()
            return

        object_map = OrderedDict()
        remote_pk_map = {
            tuple(child_mapper.primary_key_from_instance(item)):
            item for item in collection
        }
        for item in value:
            if item in object_dict:
                object_map[item] = object_dict[item]
                continue

            pk = _get_pk_from_json(child_class,
                                   child_mapper.primary_key, item,
                                   prop.local_remote_pairs, self)
            pk_incomplete = any(i is None for i in pk)

            if pk_incomplete or tuple(pk) not in remote_pk_map:
                if pk_incomplete:
                    child_obj = _polymorphic_constructor(
                        child_class, item)
                    for_update = False
                else:
                    child_obj = Session.object_session(
                        self).query(child_class).get(pk)
                    if child_obj is None:
                        if child_class.__table__._autoincrement_column is None:
                            child_obj = _polymorphic_constructor(
                                child_class, value)
                            for_update = False
                        else:
                            raise AssertionError(
                                f'Could not find object of {child_class!r}'
                                f'[{pk!r}] in database')
                    else:
                        for_update = True
                object_map[item] = child_obj
                __apply_changes(child_obj, item, object_dict, context,
                                for_update)
            else:
                child_obj = remote_pk_map[tuple(pk)]
                object_map[item] = child_obj
                __apply_changes(child_obj, item, object_dict, context,
                                True)

        for local, remote in object_map.items():
            if remote not in collection:
                collection.append(remote)

        for item in collection:
            if item not in object_map.values():
                collection.remove(item)

        # reorder if collection is ordering list
        if hasattr(collection, 'reorder') and callable(collection.reorder):
            template = list(object_map.values())
            collection.sort(key=lambda x: template.index(x))
            collection.reorder()
    else:
        # single item
        if value is None:
            attr.__set__(self, value)
        else:
            if value in object_dict:
                attr.__set__(self, object_dict[value])
            else:
                pk = _get_pk_from_json(child_class, child_mapper.primary_key,
                                       value, prop.local_remote_pairs, self)

                current_value = attr.__get__(self, None)

                # TODO
                # 1) we need to establish definitively how we tell apart bogus
                #    data from genuinely new objects; some keys may be natural?
                #    for now check nulls in PK and assume that null PKs means
                #    this is a new object
                # 2) Are there any cases when deferring this would be useful?

                pk_incomplete = any(i is None for i in pk)

                if pk_incomplete or current_value is None or \
                        child_mapper.primary_key_from_instance(current_value) \
                        != pk:
                    if pk_incomplete:
                        child_obj = _polymorphic_constructor(
                            child_class, value)
                        for_update = False
                    else:
                        child_obj = Session.object_session(
                            self).query(child_class).get(pk)
                        if child_obj is None:
                            if child_class.__table__._autoincrement_column \
                                    is None:
                                child_obj = _polymorphic_constructor(
                                    child_class, value)
                                for_update = False
                            else:
                                raise AssertionError(
                                    f'Could not find object of {child_class!r}'
                                    f'[{pk!r}] in database')
                        else:
                            for_update = True
                    __apply_changes(child_obj, value, object_dict, context,
                                    for_update)
                    attr.__set__(self, child_obj)
                else:
                    __apply_changes(current_value, value, object_dict,
                                    context, True)


def patch_sqlalchemy_base_class(base_class: type):
    base_class.apply_changes = __apply_changes



import simplejson
from .json_object import JsonObject


class JSONDecoder(simplejson.JSONDecoder):
    resolved = None
    unresolved = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, object_hook=self.custom_object_hook, **kwargs)
        self.resolved = dict()
        self.unresolved = dict()

    def custom_object_hook(self, value):
        # references will probably be resolved late
        if len(value) == 1 and "_ref" in value:
            _id = value['_ref']
            if _id in self.resolved:
                return self.resolved[_id]
            if _id in self.unresolved:
                return self.unresolved[_id]
            self.unresolved[_id] = JsonObject()
            return self.unresolved[_id]

        if '_id' in value:
            _id = value['_id']
            del value['_id']

            if _id in self.resolved:
                raise AssertionError(
                    f'two objects with the same _id in payload: {_id}')

            if _id in self.unresolved:
                ret = self.unresolved[_id]
                del self.unresolved[_id]
                ret.update(value)
            else:
                ret = JsonObject(value)
            self.resolved[_id] = ret

            return ret

        return JsonObject(value)

    def decode(self, s, *args, **kwargs):
        ret = super().decode(s, *args, **kwargs)
        if self.unresolved:
            raise AssertionError(
                'Unresolved references: ' +
                ", ".join([str(i) for i in self.unresolved.keys()]))
        self.resolved.clear()
        return ret



import re
from sqlalchemy.types import SchemaType, TypeDecorator, Enum


def EnumAttrs(name, schema=None):
    def f(cls):
        if schema is not None:
            cls.__db_schema = schema
        cls.__db_name = name
        return cls
    return f


class PythonEnum(TypeDecorator, SchemaType):
    impl = Enum

    def __init__(self, enum_class, **kw):
        if hasattr(enum_class, "__db_name") and 'name' not in kw:
            kw['name'] = getattr(enum_class, "__db_name")
        if hasattr(enum_class, "__db_schema") and 'schema' not in kw:
            kw['schema'] = getattr(enum_class, "__db_schema")
        if 'name' not in kw:
            kw['name'] = "ck%s" % re.sub('([A-Z])',
                                         lambda m: '_' + m.group(1).lower(),
                                         enum_class.__name__)
        self.impl = Enum(*(m.name for m in enum_class), **kw)
        # super().__init__(*(m.name for m in enum_class), **kw)
        self._enum_class = enum_class

    def process_bind_param(self, value, dialect):
        if isinstance(value, self._enum_class):
            return value.name
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            return self._enum_class[value]

    @property
    def python_type(self):
        return self._enum_class

    def copy(self):
        return PythonEnum(self._enum_class)

    def _set_table(self, table, column):
        self.impl._set_table(table, column)

    def __repr__(self):
        return repr(self.impl)



import simplejson
from datetime import (datetime, date, time, timedelta)
from isodate import duration_isoformat
from sqlalchemy.inspection import inspect
from sqlalchemy.ext.hybrid import HYBRID_METHOD
from sqlalchemy.ext.associationproxy import ASSOCIATION_PROXY
import base64
from enum import Enum
import uuid
from .interfaces import JsonGuardProvider
from .json_object import JsonObject, JsonOrderedObject


class JSONEncoder(simplejson.JSONEncoder):
    request = None
    obj_index = None
    base_type = None
    counter = 0
    sort = False

    def __init__(self, request=None, base_type=None, sort=False, **kwargs):
        super().__init__(encoding=None, **kwargs)
        self.request = request
        self.obj_index = dict()
        self.base_type = base_type
        self.sort = sort

    def default(self, o):
        # handle date and time format
        if isinstance(o, (datetime, date, time)):
            return o.isoformat()
        if isinstance(o, timedelta):
            return duration_isoformat(o)

        # bytes
        if type(o) == bytes:
            return base64.b64encode(o).decode('utf8')

        # enums
        if isinstance(o, Enum):
            return o.name

        # UUIDs
        if isinstance(o, uuid.UUID):
            return str(o)

        # database objects
        if self.base_type is not None and isinstance(o, self.base_type):
            mapper = inspect(type(o))
            state = inspect(o)

            composite_props = {
                prop for comp in mapper.composites for prop in comp.props}

            pk = mapper.primary_key_from_instance(o)
            pk_index = (type(o),) + tuple(pk)
            if not all([val is None for val in pk]):
                if pk_index in self.obj_index:
                    referred = self.obj_index[pk_index]
                    return dict(_ref=referred['_id'])

            # making sure all paths are explicitly loaded eliminates most of
            # the pressure to provide JSON hints at runtime; instead make sure
            # all relevant paths are loaded
            ret = JsonOrderedObject() if self.sort else JsonObject()
            descriptors = mapper.all_orm_descriptors.items()
            if self.sort:
                descriptors = sorted(descriptors)
            ret.update({
                key: getattr(o, key)
                for key, value in descriptors
                if key != '__mapper__' and key not in state.unloaded
                and value.extension_type not in (HYBRID_METHOD,
                                                 ASSOCIATION_PROXY)
                and (not hasattr(value, 'property')
                     or value.property not in composite_props)
            })
            self.counter += 1
            ret['_id'] = self.counter
            self.obj_index[pk_index] = ret
            if self.request is not None and isinstance(self.request.context,
                                                       JsonGuardProvider):
                self.request.context.guardSerialize(o, ret)
            return ret

        # end of handled types, we cannot serialize this type
        return super().default(o)



import unittest
from pyramid import testing
import transaction


class EngineTest(unittest.TestCase):
    # base for tests that require an engine
    def setUp(self):
        self.config = testing.setUp(settings={
            'sqlalchemy.url': 'sqlite:///:memory:'
        })
        from .models import includeme
        self.config.include(includeme)
        settings = self.config.get_settings()

        from ..tests.models import (
            get_engine,
            get_session_factory,
            get_tm_session
        )

        self.engine = get_engine(settings)
        session_factory = get_session_factory(self.engine)

        self.session = get_tm_session(session_factory, transaction.manager)

    def init_database(self):
        from ..tests.models import Base
        from py_liant.monkeypatch import patch_sqlalchemy_base_class
        patch_sqlalchemy_base_class(Base)
        Base.metadata.create_all(self.engine)

    def tearDown(self):
        from ..tests.models import Base

        testing.tearDown()
        transaction.abort()
        Base.metadata.drop_all(self.engine)


class TestJsonEncoderDecoder(EngineTest):

    expected_json = '''\
{
    "children": [
        {
            "data": "child value",
            "id": 1,
            "parent": {
                "_ref": 1
            },
            "parent_id": 1,
            "_id": 2
        }
    ],
    "data1": "parent value",
    "data2": "2000-01-01T00:00:00",
    "data3": "P1DT7H12M",
    "data4": "dGVzdA==",
    "data5": "type1",
    "id": 1,
    "_id": 1
}'''

    expected_json2 = '''\
{
    "children": [
        {
            "data": "child value",
            "id": 1,
            "parent_id": 1,
            "_id": 2
        }
    ],
    "data1": "parent value",
    "data2": "2000-01-01T00:00:00",
    "data3": "P1DT7H12M",
    "data4": "dGVzdA==",
    "data5": "type1",
    "id": 1,
    "_id": 1
}'''

    json_to_decode = '''\
{
    "children": [
        {
            "data": "child value",
            "id": 1,
            "parent": {
                "_ref": 1
            },
            "parent_id": 1,
            "_id": 2
        }
    ],
    "other_collection": [
        {"_ref": "custom"},
        {"_ref": 1},
        {
            "_id": "custom",
            "data": "custom object data"
        }
    ],
    "_id": 1
}'''

    def setUp(self):
        super().setUp()
        self.init_database()

        from ..tests.models import Parent, Child, ParentType
        from datetime import datetime, timedelta
        parent = Parent(data1='parent value',
                        data2=datetime(2000, 1, 1, 0, 0, 0),
                        data3=timedelta(days=1.3),
                        data4=b'test',
                        data5=ParentType.type1)
        parent.children.append(Child(data='child value'))
        self.session.add(parent)

    def test_json_serialization(self):
        from py_liant.json_encoder import JSONEncoder
        from ..tests.models import Base, Parent
        from sqlalchemy.orm import joinedload
        encoder = JSONEncoder(base_type=Base, check_circular=False,
                              indent=4 * ' ', sort=True)
        obj = self.session.query(Parent).options(
            joinedload(Parent.children)).get(1)
        info = encoder.encode(obj)
        self.assertMultiLineEqual(info, self.expected_json,
                                  "JSON encoded correctly")

    def test_json_deserialization(self):
        from py_liant.json_decoder import JSONDecoder
        from py_liant.json_object import JsonObject
        decoder = JSONDecoder()
        obj = decoder.decode(self.json_to_decode)
        self.assertIsNotNone(obj, "JSON decoded successfully")
        self.assertIsInstance(obj, JsonObject,
                              "expected JsonObject from decoder")
        self.assertFalse('_id' in obj, "_id succesfully removed from object")
        self.assertIsNotNone(obj.children, "children present")
        self.assertEqual(len(obj.children), 1, "expected one child")
        self.assertIs(obj.children[0].parent, obj,
                      "child's parent loops back on object")
        self.assertEqual(len(obj.other_collection), 3,
                         "other collection size check")
        custom_obj = obj.other_collection[0]
        self.assertEqual(custom_obj.data, "custom object data",
                         "check data in custom object")
        self.assertIs(
            obj.other_collection[2], custom_obj,
            "reference check custom object inverse/delayed resolution")

    # JSON serialization test after parent expunged;
    # we're join-loading children but parent should not show up anymore

    def test_json_serialization_after_expunge(self):
        from py_liant.json_encoder import JSONEncoder
        from ..tests.models import Base, Parent
        from sqlalchemy.orm import joinedload
        encoder = JSONEncoder(base_type=Base, check_circular=False,
                              indent=4 * ' ', sort=True)
        obj = self.session.query(Parent).get(1)
        self.session.expunge(obj)
        obj = self.session.query(Parent).options(joinedload(Parent.children))\
            .get(1)
        info = encoder.encode(obj)
        self.assertMultiLineEqual(info, self.expected_json2,
                                  "JSON encoded correctly")

    # test failure mode: unresolved reference
    def test_json_unresolved(self):
        from py_liant.json_decoder import JSONDecoder
        decoder = JSONDecoder()
        self.assertRaisesRegex(AssertionError, "Unresolved references",
                               decoder.decode, '{"_ref": 1}')

    # test failure mode: _id collision
    def test_json_id_collision(self):
        from py_liant.json_decoder import JSONDecoder
        decoder = JSONDecoder()
        self.assertRaisesRegex(
            AssertionError, "two objects with the same _id", decoder.decode,
            '{"items": [{"_id": 1}, {"_id": 1}]}')

    def test_json_uuid_value(self):
        from py_liant.json_encoder import JSONEncoder
        import uuid
        encoder = JSONEncoder()
        info = encoder.encode({'uuid': uuid.uuid5(uuid.NAMESPACE_DNS,
                                                  "test.com")})
        self.assertEqual(
            info, '{"uuid": "1c39b279-6010-55d9-b677-859bffab8081"}',
            "check UUID encoding")


class TestApplyChanges(EngineTest):
    def setUp(self):
        super().setUp()
        self.init_database()

        from ..tests.models import Parent, Child, ParentType
        from datetime import datetime, timedelta
        parent = Parent(data1='parent value',
                        data2=datetime(2000, 1, 1, 0, 0, 0),
                        data3=timedelta(days=1.3),
                        data4=b'test',
                        data5=ParentType.type1)
        parent.children.append(Child(data='child value'))
        self.session.add(parent)

    def test_monkeypatch(self):
        from ..tests.models import Parent
        obj = Parent()
        self.assertTrue(callable(obj.apply_changes), "patch method exists")

    def test_property_change(self):
        from py_liant.json_decoder import JSONDecoder
        from ..tests.models import Parent
        decoder = JSONDecoder()

        obj = self.session.query(Parent).get(1)
        data = decoder.decode('''{"data1": "changed value"}''')
        obj.apply_changes(data)
        self.assertEqual(obj.data1, "changed value",
                         "Parent.data correctly changed")

    def test_deep_property_change(self):
        from py_liant.json_decoder import JSONDecoder
        from ..tests.models import Parent
        decoder = JSONDecoder()

        obj = self.session.query(Parent).get(1)
        obj_child = obj.children[0]

        data = decoder.decode('''\
            {
                "children": [
                    { "id": 1, "data": "child changed value" },
                    { "data": "new child value" }
                ]
            }
            ''')
        obj.apply_changes(data)
        self.assertIs(obj.children[0], obj_child, "first child preserved")
        self.assertEqual(obj_child.data, "child changed value")
        self.assertEqual(len(obj.children), 2, "two children present now")
        obj_child2 = obj.children[1]
        self.assertIsNone(obj_child2.id, "second child is transient")
        self.assertEqual(obj_child2.data, "new child value",
                         "second child correct value")



from abc import ABC, abstractmethod


class JsonGuardProvider(ABC):
    @abstractmethod
    def guardUpdate(self, obj, data, for_update=True):
        pass

    @abstractmethod
    def guardHints(self, cls, hints):
        pass

    @abstractmethod
    def guardSerialize(self, obj, value):
        pass

    @abstractmethod
    def guardDrilldown(self, prop):
        pass



from sqlalchemy.interfaces import PoolListener
import logging

sa_logger = logging.getLogger('sqlalchemy')


class SearchPathSetter(PoolListener):
    '''
    Dynamically sets the search path on connections checked out from a pool.
    '''

    def __init__(self, search_path='public'):
        self.search_path = search_path

    @staticmethod
    def quote_schema(dialect, schema):
        return dialect.identifier_preparer.quote_schema(schema, False)

    def checkout(self, dbapi_con, con_record, con_proxy):
        dialect = con_proxy._pool._dialect
        search_path = [
            self.quote_schema(dialect, _.strip())
            for _ in self.search_path.split(',')
            if _.strip()
        ]
        if 'public' not in search_path:
            search_path.append('public')
        cursor = dbapi_con.cursor()
        statement = "SET search_path TO %s;" % ', '.join(search_path)
        sa_logger.info(statement)
        cursor.execute(statement)
        dbapi_con.commit()
        cursor.close()



from typing import Dict
from pyparsing import ParseException
from pyramid.request import Request
from pyramid.httpexceptions import (
    HTTPNotFound, HTTPServerError, HTTPConflict, HTTPOk)
from sqlalchemy.orm.util import AliasedClass
from sqlalchemy.orm.exc import (NoResultFound, StaleDataError)
from sqlalchemy.exc import InvalidRequestError
from sqlalchemy.orm.base import NOT_EXTENSION
from sqlalchemy.orm import ColumnProperty, RelationshipProperty, Mapper
from sqlalchemy import String, orm, and_
from sqlalchemy.inspection import inspect
from pyramid.settings import asbool
import transaction


from .json_encoder import JSONEncoder
from .json_decoder import JSONDecoder
from .monkeypatch import coerce_value, patch_sqlalchemy_base_class
from .parser import route_parser, hints_parser
from .interfaces import JsonGuardProvider


# Returns a renderer for pyramid; base_type (SQLAlchemy declarative base) is
# needed to detect sqlalchemy object instances
# Sample usage:
#     config.add_renderer('json', pyramid_json_renderer_factory(Base))

def pyramid_json_renderer_factory(base_type=None, wsgi_iter=False,
                                  separators=(',', ':')):
    def _json_renderer(info):
        def _render(value, system):
            request = system.get('request')
            if request is not None:
                # this is optimal as we will stream our JSON from the iterator
                response = request.response
                response.content_type = 'application/json'
                response.charset = 'utf8'

                json_encoder = JSONEncoder(request, base_type=base_type,
                                           separators=separators,
                                           check_circular=False)

                # solution 1: write to stream from renderer
                if not wsgi_iter:
                    for chunk in json_encoder.iterencode(value):
                        response.write(chunk)
                    return None

                # solution 2: provide iterable to wsgi layer
                else:
                    def _iterencode(val):
                        for chunk in json_encoder.iterencode(val):
                            yield chunk.encode()

                    response.app_iter = _iterencode(value)
                    return None
            else:
                # fallback for direct calls?
                json_encoder = JSONEncoder(base_type=base_type,
                                           separators=separators)
                return json_encoder.encode(value)
        return _render
    return _json_renderer


# Sample usage:
#     config.add_request_method(pyramid_json_decoder, 'json', reify=True)
def pyramid_json_decoder(request):
    return JSONDecoder(encoding=request.charset).decode(request.body)


# exception to be thrown when a
class VersionCheckError(RuntimeError):
    pass


# Extend this and decorate accordingly; methods list, get, insert, update,
# delete should be decorated with pyramid @view_config
class CRUDView(object):
    filters = dict()
    accept_order = dict()
    # json_hints = dict()
    # update_hints = dict()
    context = None
    target_type = object
    target_name = "object"
    request = None
    update_lock = False
    # this means accept_order will be rewritten
    use_subquery_after_filter = False

    def __init__(self, request):
        """:type request: Request"""
        self.request = request
        self.context = request.context
        # self.context._json_hints = self.json_hints
        assert(isinstance(self.request, Request))

    @property
    def identity_filter(self):
        return False

    def sanitize_input(self):
        return self.request.json[self.target_name]

    @property
    def context_filter(self):
        return True

    def get_one_from_query(self, query):
        return query.filter(self.context_filter, self.identity_filter).one()

    def get_by_id(self, update_lock=False):
        try:
            query = self.get_identity_base()
            if update_lock:
                # (partially a) SQLAlchemy bug: Postgres doesn't allow
                # qualified table names in FOR UPDATE OF statements, so
                # whenever targeted locks are required an alias is necessary,
                # for us at least since we use schemas for all tables
                if isinstance(self.target_type, AliasedClass):
                    query = query.with_for_update(of=self.target_type)
                else:
                    query = query.with_for_update()
            return self.get_one_from_query(query)
        except NoResultFound:
            raise HTTPNotFound()

    def get_query_filters(self, exclude=None):
        tmp = [self.filters[key](self.request.GET[key]) for key in self.filters
               if key in self.request.GET and
               (exclude is None or key not in exclude)]
        return [i for i in tmp if not isinstance(i, (list, tuple))
                and i is not None] + \
               [i for j in tmp if isinstance(j, (list, tuple))
                and j is not None for i in j if i is not None] + \
            self.runtime_filters()

    def runtime_filters(self):
        return []

    @property
    def query_filters(self):
        return self.get_query_filters()

    @property
    def order_clauses(self):
        if 'order' not in self.request.GET or self.request.GET['order'] == "":
            return None
        orders = [(item[:-5], True) if item.endswith(' desc')
                  else(item, False)
                  for item in self.request.GET['order'].split(",")]
        if any([order[0] not in self.accept_order for order in orders]):
            invalid_order = [order[0] for order in orders
                             if order[0] not in self.accept_order]
            raise HTTPServerError(
                f'not implemented, order by {", ".join(invalid_order)}'
            )

        selected = [(self.accept_order[order[0]], order[1])
                    for order in orders]
        for item in selected:
            if isinstance(item[0], list):
                for elem in item[0]:
                    yield elem.desc() if item[1] else elem
            else:
                yield item[0].desc() if item[1] else item[0]

    @property
    def pager_slice(self):
        if 'pageSize' not in self.request.GET:
            return False
        page_size = int(self.request.GET['pageSize'])
        if page_size <= 0:
            return False
        page = int(self.request.GET['page']
                   ) if 'page' in self.request.GET else 0
        return slice(page * page_size, page * page_size + page_size)

    def get_base_query(self):
        return self.request.dbsession.query(self.target_type)

    def get_identity_base(self):
        return self.get_base_query()

    def get_list_base(self):
        return self.get_base_query()

    def get_search_results(self, query=None):
        if query is None:
            query = self.get_list_base().filter(self.context_filter)
        query = query.filter(*self.query_filters)
        pager = self.pager_slice
        count = query.count()

        if self.use_subquery_after_filter:
            query = query.subquery()
            self.accept_order = dict(query.c)
            query = self.request.dbsession.query(
                *query.c)

        order_clauses = self.order_clauses
        if order_clauses is not None:
            # reset and apply order_by
            query = query.order_by(None).order_by(*order_clauses)
        return query[pager] if pager else query.all(), count

    def get(self):
        return {self.target_name: self.get_by_id()}

    def list(self):
        items, count = self.get_search_results()
        return dict(items=items, total=count)

    def apply_changes(self, obj, data, for_update=True):
        obj.apply_changes(data, context=self.context, for_update=for_update)

    def update(self):
        try:
            with transaction.manager, self.request.dbsession.no_autoflush:
                old = self.get_by_id(update_lock=self.update_lock)
                values = self.sanitize_input()
                try:
                    self.apply_changes(old, values, True)
                except VersionCheckError as ex:
                    raise HTTPConflict(str(ex))
        except StaleDataError as ex:
            raise HTTPConflict(str(ex))
        return self.request.dbsession.merge(old)

    def insert(self):
        obj = self.target_type()
        values = self.sanitize_input()
        with transaction.manager:
            self.request.dbsession.add(obj)
            self.apply_changes(obj, values, False)
        obj = self.request.dbsession.merge(obj)
        return {self.target_name: obj}

    def delete(self):
        with transaction.manager:
            old = self.get_by_id()
            self.request.dbsession.delete(old)
        return HTTPOk()

    @classmethod
    def auto_fields(cls, target):
        return [item for item in inspect(target).all_orm_descriptors.values()
                if item.is_attribute and item.extension_type == NOT_EXTENSION
                and isinstance(item.property, ColumnProperty)]

    @classmethod
    def auto_filters(cls, target=None, prefix=None):
        if target is None:
            target = cls.target_type

        def coerce_func(x, attr):
            return coerce_value(target, attr.property.columns[0], x, False)

        ret = dict()

        for item in cls.auto_fields(target):
            key = prefix + item.key if prefix is not None else item.key
            ret[key] = lambda x, attr=item: attr == coerce_func(x, attr)
            if isinstance(item.property.columns[0].type, String):
                ret[f'{key}_like'] = \
                    lambda x, attr=item: attr.ilike('%' + x + '%')
            ret[f'{key}_gt'] = \
                lambda x, attr=item: attr > coerce_func(x, attr)
            ret[f'{key}_ge'] = \
                lambda x, attr=item: attr >= coerce_func(x, attr)
            ret[f'{key}_lt'] = \
                lambda x, attr=item: attr < coerce_func(x, attr)
            ret[f'{key}_le'] = \
                lambda x, attr=item: attr <= coerce_func(x, attr)
            ret[f'{key}_isnull'] = \
                lambda x, attr=item: attr.is_(None) if asbool(x) else \
                attr.isnot(None)
        return ret

    @classmethod
    def auto_order(cls, target=None, prefix=None):
        if target is None:
            target = cls.target_type
        ret = dict()
        for attr in cls.auto_fields(target):
            key = prefix + attr.key if prefix is not None else attr.key
            ret[key] = attr
        return ret


class ConvertMatchdictPredicate:
    def __init__(self, args, config):
        assert len(args) >= 1 and (type(args[0]) == tuple or
                                   callable(args[0]))
        self.args = args

    def text(self):
        return f'convert_matchdict={self.args!r}'

    phash = text

    def __call__(self, context, request):
        args = self.args if type(self.args[0]) == tuple else(self.args,)
        match = request.matchdict
        for argList in args:
            argType = argList[0]
            assert callable(argType)
            for arg in (argList[1:]):
                if arg in match:
                    try:
                        match[arg] = argType(match[arg])
                    except (TypeError, ValueError):
                        return False
        return True


def _get_assert_one(q):
    return q.one()


def _get_by_pkey(pkey):
    def _impl(q):
        return q.get(pkey)
    return _impl

def _get_by_combinedfilter(pkey_filter):
    def _impl(q):
        return q.filter(*pkey_filter).one()
    return _impl

# TODO: reverse order for negative indexes (?)
def _get_by_index(index):
    def _impl(q):
        ret = q[index:index + 1]
        if len(ret) == 0:
            raise NoResultFound()
        return ret[0]
    return _impl


class CatchallTarget:
    _cls = None
    profiles = dict()
    filters = []

    def __init__(self, cls, profiles=dict(), filters=None):
        self._cls, self.profiles, self.filters = cls, profiles, \
            filters

    def __repr__(self):
        return f"(cls:{self._cls!r}, profiles:{self.profiles!r}, filters:{self.filters!r}"


class CatchallPredicate:
    targets: Dict[str, CatchallTarget] = None

    def __init__(self, targets, config):
        def _adapt(obj) -> CatchallTarget:
            if isinstance(obj, CatchallTarget):
                return obj
            if type(obj) not in (tuple, dict, list):
                return CatchallTarget(obj)
            if type(obj) is dict:
                return CatchallTarget(**obj)
            return CatchallTarget(*obj)

        self.targets = {
            key: _adapt(obj) for key, obj in targets.items()
        }

    def text(self):
        return f'catchall={self.targets!r}'

    phash = text

    def __call__(self, context, request):
        match = request.matchdict

        # convert target to type
        if 'catchall' not in match:
            return False

        try:
            route = route_parser.parseString(match['catchall'], True)
        except ParseException:
            return False

        if route['verb'] not in self.targets:
            return False

        target = self.targets[route['verb']]
        getter = None

        insp = inspect(target._cls)
        if 'cast' in route:
            if insp.polymorphic_on is None:
                return False
            try:
                value = coerce_value(target._cls, insp.polymorphic_on,
                                     route['cast'])
            except ValueError:
                return False
            if value not in insp.polymorphic_map:
                return False
            insp = insp.polymorphic_map[value]
            target = CatchallTarget(insp.class_, target.profiles,
                                    target.filters)

        query = request.dbsession.query(target._cls)

        if target.filters is not None:
            filters = target.filters
            if callable(filters):
                filters = filters(request)
            if type(filters) not in (list, tuple):
                filters = (filters,)

            query = query.filter(*filters)

        # convert pkey
        if 'pkey' in route:
            pkey = route['pkey']
            if len(pkey) != len(insp.primary_key):
                return False
            try:
                pkey = tuple(coerce_value(target._cls, col, val)
                             for col, val in zip(insp.primary_key, pkey))
            except ValueError:
                return False
            if query.whereclause is None:
                # no implicit/context filters
                getter = _get_by_pkey(pkey)
            else:
                filters = map(
                    lambda x: x[0]==x[1], 
                    zip(insp.primary_key, pkey)
                    )
                getter = _get_by_combinedfilter(filters)

        if 'drilldown' in route:
            # cannot drilldown property if result is list
            if getter is None:
                return False

            drilldown = route['drilldown']
            try:
                prop = insp.get_property(drilldown)
            except InvalidRequestError:
                return False

            if isinstance(context, JsonGuardProvider):
                if not context.guardDrilldown(prop.class_attribute):
                    return False

            if not isinstance(prop, RelationshipProperty):
                return False

            # target switch in drilldown
            target = CatchallTarget(prop.entity.class_)

            if prop.lazy == 'dynamic':
                if 'drilldown_cast' in route:
                    # we cannot apply polymorphic casting here
                    return False
                # special drilldown love for dynamic props
                query = getattr(getter(query), prop.key)
                # getter no longer useful
                getter = None
            else:
                # manually construct drilldown query for non-dynamic props
                pkey_filter = (col == val for col, val in
                               zip(insp.primary_key, pkey))
                if 'drilldown_cast' in route:
                    insp = inspect(target._cls)
                    if insp.polymorphic_on is None:
                        return False
                    try:
                        value = coerce_value(target._cls, insp.polymorphic_on,
                                             route['drilldown_cast'])
                    except ValueError:
                        return False
                    if value not in insp.polymorphic_map:
                        return False
                    insp = insp.polymorphic_map[value]
                    target = CatchallTarget(insp.class_)

                query = request.dbsession.query(target._cls) \
                    .select_from(prop.parent) \
                    .join(prop.class_attribute) \
                    .filter(and_(*pkey_filter))
                getter = None

            if not prop.uselist:
                # target is single item fk
                getter = _get_assert_one

        if 'slice' in route:
            # cannot slice if pkey or single item drilldown encountered
            if getter is not None:
                return False

            slicer = route['slice']
            if 'index' in slicer:
                getter = _get_by_index(int(slicer['index']))
            else:
                match['slicer'] = slicer

        profile = None
        if 'profile' in route:
            if route['profile'] not in target.profiles:
                return False
            profile = target.profiles[route['profile']]
            profile = hints_parser.parseString(profile, True)
            hints = self.get_hints(profile, target._cls, context=context)
            query = query.options(*hints)

        # decode hints
        if 'hints' in route:
            try:
                hints = self.get_hints(route['hints'], target._cls,
                                       context=context)
                query = query.options(*hints)
            except AssertionError:
                return False

        match['query'] = query
        match['getter'] = getter

        return True

    @staticmethod
    def get_hints(value, cls, base=orm, context=None):
        # hints structure: [atom(,atom)+]
        # atom is one of:
        #  +field_name: undefer field (include)
        #  -field_name: defer field (exclude)
        #  TODO if field_name is "ALL", defer or undefer all fields
        #   (DOES NOT AFFECT COLLECTIONS, see below)
        #  *collection(-x)?(\(hints\))?:
        #   eager load collection (-x marks subquery laod preference);
        #   hints in parantheses recursively apply to elements of
        #   the collection

        if type(cls) is Mapper:
            # caller passed us a mapper (drilldown)
            insp = cls
            cls = insp.class_
        else:
            insp = inspect(cls)

        hints = {}
        ret = []
        for item in value:
            op = item['op']
            if op == '!':
                if insp.polymorphic_on is None:
                    raise AssertionError("invalid cast")
                try:
                    _type = coerce_value(cls, insp.polymorphic_on,
                                         item['type'])
                except ValueError:
                    raise AssertionError("invalid cast")
                if _type not in insp.polymorphic_map:
                    raise AssertionError("invalid cast")

                ret.extend(
                    CatchallPredicate.get_hints(
                        item['children'],
                        insp.polymorphic_map[_type],
                        base, context=context
                    ))
                continue

            name = item['name']
            prop = insp.get_property(name)
            hints[prop.class_attribute] = (op, item.get('children'))
        if isinstance(context, JsonGuardProvider):
            context.guardHints(cls, hints)
        for attr, (op, children) in hints.items():
            prop = attr.prop
            hint = None
            if isinstance(prop, ColumnProperty):
                if op == '+':
                    hint = base.undefer(attr)
                elif op == '-':
                    hint = base.defer(attr)
                else:
                    raise AssertionError("invalid op")
            elif isinstance(prop, RelationshipProperty):
                if op == '-':
                    hint = base.lazyload(attr)
                else:
                    if op == '*':
                        hint = base.selectinload(attr)
                    else:
                        hint = base.joinedload(attr)
                    if children:
                        CatchallPredicate.get_hints(children, attr.entity,
                                                    hint, context=context)
            ret.append(hint)
        return ret


# base class to decorate with a CatchAllPredicate
class CatchallView(CRUDView):
    context = None
    request = None
    query = None
    getter = None
    slicer = None

    def __init__(self, request):
        super().__init__(request)
        self.query = self.request.matchdict['query']
        self.getter = self.request.matchdict.get('getter')

        self.target_type = self.query.column_descriptions[0]['type']
        self.target_name = self.target_type.__table__.name
        self.slicer = self.request.matchdict.get('slicer')

        self.filters = self.auto_filters(target=self.target_type)
        self.accept_order = self.auto_order(target=self.target_type)

    def get_base_query(self):
        return self.query

    def get_one_from_query(self, query):
        if self.getter is None:
            raise HTTPNotFound()
        query = query.filter(*self.query_filters)
        order_clauses = self.order_clauses
        if order_clauses is not None:
            query = query.order_by(None).order_by(*order_clauses)
        return self.getter(query)

    def process(self):
        if self.request.method == 'GET':
            if self.getter is not None:
                return self.get()
            return self.list()
        elif self.request.method == 'POST':
            if self.getter is not None:
                return self.update()
            return self.insert()
        elif self.request.method == 'DELETE':
            if self.getter is None:
                raise HTTPNotFound()
            return self.delete()
        raise HTTPNotFound()

    @ property
    def pager_slice(self):
        if self.slicer is not None:
            return slice(int(self.slicer['start']),
                         int(self.slicer['stop']))
        return super().pager_slice


def includeme_factory(base_class=None, config_json=True, add_predicates=True,
                      wsgi_iter=False, separators=(',', ':')):
    def includeme(config):
        if config_json and base_class is not None:
            config.add_renderer(
                'json',
                pyramid_json_renderer_factory(
                    base_class, wsgi_iter=wsgi_iter,
                    separators=separators))
            config.add_request_method(pyramid_json_decoder, 'json', reify=True)
        if add_predicates:
            config.add_view_predicate(
                'convert_matchdict', ConvertMatchdictPredicate)
            config.add_view_predicate('catchall', CatchallPredicate)
        if base_class is not None:
            patch_sqlalchemy_base_class(base_class)

    return includeme
